@@ -33,7 +33,6 @@
 #include "mpi.h"
 #include "cipher.h"
 #include "pubkey-internal.h"
-#include "const-time.h"
 
 
 typedef struct
@@ -178,49 +177,6 @@ test_keys (RSA_secret_key *sk, unsigned int nbits)
   return result;
 }
 
-static int
-test_keys_fips (gcry_sexp_t skey)
-{
-  int result = -1; /* Default to failure.  */
-  char plaintext[128];
-  gcry_sexp_t sig = NULL;
-  const char *data_tmpl = "(data (flags pkcs1) (hash %s %b))";
-  gcry_md_hd_t hd = NULL;
-  int ec;
-
-  /* Create a random plaintext.  */
-  _gcry_randomize (plaintext, sizeof plaintext, GCRY_WEAK_RANDOM);
-
-  /* Open MD context and feed the random data in */
-  ec = _gcry_md_open (&hd, GCRY_MD_SHA256, 0);
-  if (ec)
-    goto leave;
-  _gcry_md_write (hd, plaintext, sizeof(plaintext));
-
-  /* Use the RSA secret function to create a signature of the plaintext.  */
-  ec = _gcry_pk_sign_md (&sig, data_tmpl, hd, skey, NULL);
-  if (ec)
-    goto leave;
-
-  /* Use the RSA public function to verify this signature.  */
-  ec = _gcry_pk_verify_md (sig, data_tmpl, hd, skey, NULL);
-  if (ec)
-    goto leave;
-
-  /* Modify the data and check that the signing fails.  */
-  _gcry_md_reset(hd);
-  plaintext[sizeof plaintext / 2] ^= 1;
-  _gcry_md_write (hd, plaintext, sizeof(plaintext));
-  ec = _gcry_pk_verify_md (sig, data_tmpl, hd, skey, NULL);
-  if (ec != GPG_ERR_BAD_SIGNATURE)
-    goto leave; /* Signature verification worked on modified data  */
-
-  result = 0; /* All tests succeeded.  */
- leave:
-  sexp_release (sig);
-  _gcry_md_close (hd);
-  return result;
-}
 
 /* Callback used by the prime generation to test whether the exponent
    is suitable. Returns 0 if the test has been passed. */
@@ -264,6 +220,14 @@ generate_std (RSA_secret_key *sk, unsigned int nbits, unsigned long use_e,
   gcry_mpi_t f;
   gcry_random_level_t random_level;
 
+  if (fips_mode ())
+    {
+      if (nbits < 1024)
+        return GPG_ERR_INV_VALUE;
+      if (transient_key)
+        return GPG_ERR_INV_VALUE;
+    }
+
   /* The random quality depends on the transient_key flag.  */
   random_level = transient_key ? GCRY_STRONG_RANDOM : GCRY_VERY_STRONG_RANDOM;
 
@@ -392,39 +356,6 @@ generate_std (RSA_secret_key *sk, unsigned int nbits, unsigned long use_e,
 }
 
 
-/* Check the RSA key length is acceptable for key generation or usage */
-static gpg_err_code_t
-rsa_check_keysize (unsigned int nbits)
-{
-  if (fips_mode () && nbits < 2048)
-    return GPG_ERR_INV_VALUE;
-
-  return GPG_ERR_NO_ERROR;
-}
-
-
-/* Check the RSA key length is acceptable for signature verification
- *
- * FIPS allows signature verification with RSA keys of size
- * 1024, 1280, 1536 and 1792 in legacy mode, but this is up to the
- * calling application to decide if the signature is legacy and
- * should be accepted.
- */
-static gpg_err_code_t
-rsa_check_verify_keysize (unsigned int nbits)
-{
-  if (fips_mode ())
-    {
-      if ((nbits >= 1024 && (nbits % 256) == 0) || nbits >= 2048)
-        return GPG_ERR_NO_ERROR;
-
-      return GPG_ERR_INV_VALUE;
-    }
-
-  return GPG_ERR_NO_ERROR;
-}
-
-
 /****************
  * Generate a key pair with a key of size NBITS.
  * USE_E = 0 let Libcgrypt decide what exponent to use.
@@ -454,16 +385,12 @@ generate_fips (RSA_secret_key *sk, unsigned int nbits, unsigned long use_e,
   unsigned int pbits = nbits/2;
   unsigned int i;
   int pqswitch;
-  gpg_err_code_t ec;
+  gpg_err_code_t ec = GPG_ERR_NO_PRIME;
 
-  if (nbits <= 1024 || (nbits & 0x1FF))
+  if (nbits < 1024 || (nbits & 0x1FF))
     return GPG_ERR_INV_VALUE;
-  ec = rsa_check_keysize (nbits);
-  if (ec)
-    return ec;
-
-  /* Set default error code.  */
-  ec = GPG_ERR_NO_PRIME;
+  if (_gcry_enforced_fips_mode() && nbits != 2048 && nbits != 3072)
+      return GPG_ERR_INV_VALUE;
 
   /* The random quality depends on the transient_key flag.  */
   random_level = transient_key ? GCRY_STRONG_RANDOM : GCRY_VERY_STRONG_RANDOM;
@@ -542,13 +469,12 @@ generate_fips (RSA_secret_key *sk, unsigned int nbits, unsigned long use_e,
 
  retry:
   /* generate p and q */
-  for (i = 0; i < 10 * pbits; i++)
+  for (i = 0; i < 5 * pbits; i++)
     {
     ploop:
       if (!testparms)
         {
           _gcry_mpi_randomize (p, pbits, random_level);
-          mpi_set_bit (p, 0);
         }
       if (mpi_cmp (p, minp) < 0)
         {
@@ -572,16 +498,15 @@ generate_fips (RSA_secret_key *sk, unsigned int nbits, unsigned long use_e,
       else if (testparms)
         goto err;
     }
-  if (i >= 10 * pbits)
+  if (i >= 5 * pbits)
     goto err;
 
-  for (i = 0; i < 20 * pbits; i++)
+  for (i = 0; i < 5 * pbits; i++)
     {
     qloop:
       if (!testparms)
         {
           _gcry_mpi_randomize (q, pbits, random_level);
-          mpi_set_bit (q, 0);
         }
       if (mpi_cmp (q, minp) < 0)
         {
@@ -621,7 +546,7 @@ generate_fips (RSA_secret_key *sk, unsigned int nbits, unsigned long use_e,
       else if (testparms)
         goto err;
     }
-  if (i >= 20 * pbits)
+  if (i >= 5 * pbits)
     goto err;
 
   if (testparms)
@@ -663,7 +588,7 @@ generate_fips (RSA_secret_key *sk, unsigned int nbits, unsigned long use_e,
       mpi_invm (u, p, q );
     }
 
-  ec = 0; /* Success.  */
+  ec = 0;
 
   if (DBG_CIPHER)
     {
@@ -691,7 +616,8 @@ generate_fips (RSA_secret_key *sk, unsigned int nbits, unsigned long use_e,
   sk->d = d;
   sk->u = u;
 
-  if (ec)
+  /* Now we can test our keys. */
+  if (ec || (!testparms && test_keys (sk, nbits - 64)))
     {
       _gcry_mpi_release (sk->n); sk->n = NULL;
       _gcry_mpi_release (sk->e); sk->e = NULL;
@@ -699,6 +625,11 @@ generate_fips (RSA_secret_key *sk, unsigned int nbits, unsigned long use_e,
       _gcry_mpi_release (sk->q); sk->q = NULL;
       _gcry_mpi_release (sk->d); sk->d = NULL;
       _gcry_mpi_release (sk->u); sk->u = NULL;
+      if (!ec)
+        {
+          fips_signal_error ("self-test after key generation failed");
+          return GPG_ERR_SELFTEST_FAILED;
+        }
     }
 
   return ec;
@@ -765,9 +696,7 @@ generate_x931 (RSA_secret_key *sk, unsigned int nbits, unsigned long e_value,
 
   *swapped = 0;
 
-  if (e_value == 0)        /* 65537 is the libgcrypt's selection. */
-    e_value = 65537;
-  else if (e_value == 1)   /* Alias for a secure value. */
+  if (e_value == 1)   /* Alias for a secure value. */
     e_value = 65537;
 
   /* Point 1 of section 4.1:  k = 1024 + 256s with S >= 0  */
@@ -1219,7 +1148,6 @@ rsa_generate (const gcry_sexp_t genparms, gcry_sexp_t *r_skey)
   int flags = 0;
   gcry_sexp_t l1;
   gcry_sexp_t swap_info = NULL;
-  int testparms = 0;
 
   memset (&sk, 0, sizeof sk);
 
@@ -1257,11 +1185,6 @@ rsa_generate (const gcry_sexp_t genparms, gcry_sexp_t *r_skey)
   if (deriveparms || (flags & PUBKEY_FLAG_USE_X931))
     {
       int swapped;
-      if (fips_mode ())
-        {
-          sexp_release (deriveparms);
-          return GPG_ERR_INV_SEXP;
-        }
       ec = generate_x931 (&sk, nbits, evalue, deriveparms, &swapped);
       sexp_release (deriveparms);
       if (!ec && swapped)
@@ -1281,11 +1204,9 @@ rsa_generate (const gcry_sexp_t genparms, gcry_sexp_t *r_skey)
         }
       deriveparms = (genparms? sexp_find_token (genparms, "test-parms", 0)
                      /**/    : NULL);
-      if (deriveparms)
-        testparms = 1;
 
       /* Generate.  */
-      if (deriveparms || fips_mode ())
+      if (deriveparms || fips_mode())
         {
           ec = generate_fips (&sk, nbits, evalue, deriveparms,
                               !!(flags & PUBKEY_FLAG_TRANSIENT_KEY));
@@ -1320,13 +1241,6 @@ rsa_generate (const gcry_sexp_t genparms, gcry_sexp_t *r_skey)
   mpi_free (sk.u);
   sexp_release (swap_info);
 
-  if (!ec && !testparms && fips_mode () && test_keys_fips (*r_skey))
-    {
-      sexp_release (*r_skey); *r_skey = NULL;
-      fips_signal_error ("self-test after key generation failed");
-      return GPG_ERR_SELFTEST_FAILED;
-    }
-
   return ec;
 }
 
@@ -1368,13 +1282,9 @@ rsa_encrypt (gcry_sexp_t *r_ciph, gcry_sexp_t s_data, gcry_sexp_t keyparms)
   gcry_mpi_t data = NULL;
   RSA_public_key pk = {NULL, NULL};
   gcry_mpi_t ciph = NULL;
-  unsigned int nbits = rsa_get_nbits (keyparms);
 
-  rc = rsa_check_keysize (nbits);
-  if (rc)
-    return rc;
-
-  _gcry_pk_util_init_encoding_ctx (&ctx, PUBKEY_OP_ENCRYPT, nbits);
+  _gcry_pk_util_init_encoding_ctx (&ctx, PUBKEY_OP_ENCRYPT,
+                                   rsa_get_nbits (keyparms));
 
   /* Extract the data.  */
   rc = _gcry_pk_util_data_to_mpi (s_data, &data, &ctx);
@@ -1436,7 +1346,7 @@ static gcry_err_code_t
 rsa_decrypt (gcry_sexp_t *r_plain, gcry_sexp_t s_data, gcry_sexp_t keyparms)
 
 {
-  gpg_err_code_t rc, rc_sexp;
+  gpg_err_code_t rc;
   struct pk_encoding_ctx ctx;
   gcry_sexp_t l1 = NULL;
   gcry_mpi_t data = NULL;
@@ -1444,15 +1354,9 @@ rsa_decrypt (gcry_sexp_t *r_plain, gcry_sexp_t s_data, gcry_sexp_t keyparms)
   gcry_mpi_t plain = NULL;
   unsigned char *unpad = NULL;
   size_t unpadlen = 0;
-  unsigned int nbits = rsa_get_nbits (keyparms);
-  gcry_sexp_t result = NULL;
-  gcry_sexp_t dummy = NULL;
 
-  rc = rsa_check_keysize (nbits);
-  if (rc)
-    return rc;
-
-  _gcry_pk_util_init_encoding_ctx (&ctx, PUBKEY_OP_DECRYPT, nbits);
+  _gcry_pk_util_init_encoding_ctx (&ctx, PUBKEY_OP_DECRYPT,
+                                   rsa_get_nbits (keyparms));
 
   /* Extract the data.  */
   rc = _gcry_pk_util_preparse_encval (s_data, rsa_names, &l1, &ctx);
@@ -1495,7 +1399,7 @@ rsa_decrypt (gcry_sexp_t *r_plain, gcry_sexp_t s_data, gcry_sexp_t keyparms)
   mpi_fdiv_r (data, data, sk.n);
 
   /* Allocate MPI for the plaintext.  */
-  plain = mpi_snew (nbits);
+  plain = mpi_snew (ctx.nbits);
 
   /* We use blinding by default to mitigate timing attacks which can
      be practically mounted over the network as shown by Brumley and
@@ -1503,7 +1407,7 @@ rsa_decrypt (gcry_sexp_t *r_plain, gcry_sexp_t s_data, gcry_sexp_t keyparms)
   if ((ctx.flags & PUBKEY_FLAG_NO_BLINDING))
     secret (plain, data, &sk);
   else
-    secret_blinded (plain, data, &sk, nbits);
+    secret_blinded (plain, data, &sk, ctx.nbits);
 
   if (DBG_CIPHER)
     log_printmpi ("rsa_decrypt  res", plain);
@@ -1512,29 +1416,21 @@ rsa_decrypt (gcry_sexp_t *r_plain, gcry_sexp_t s_data, gcry_sexp_t keyparms)
   switch (ctx.encoding)
     {
     case PUBKEY_ENC_PKCS1:
-      rc = _gcry_rsa_pkcs1_decode_for_enc (&unpad, &unpadlen, nbits, plain);
+      rc = _gcry_rsa_pkcs1_decode_for_enc (&unpad, &unpadlen, ctx.nbits, plain);
       mpi_free (plain);
       plain = NULL;
-      rc_sexp = sexp_build (&result, NULL, "(value %b)", (int)unpadlen, unpad);
-      *r_plain = sexp_null_cond (result, ct_is_not_zero (rc));
-      dummy = sexp_null_cond (result, ct_is_zero (rc));
-      sexp_release (dummy);
-      rc = ct_ulong_select (rc_sexp, rc,
-			    ct_is_zero (rc) & ct_is_not_zero (rc_sexp));
+      if (!rc)
+        rc = sexp_build (r_plain, NULL, "(value %b)", (int)unpadlen, unpad);
       break;
 
     case PUBKEY_ENC_OAEP:
       rc = _gcry_rsa_oaep_decode (&unpad, &unpadlen,
-                                  nbits, ctx.hash_algo,
+                                  ctx.nbits, ctx.hash_algo,
                                   plain, ctx.label, ctx.labellen);
       mpi_free (plain);
       plain = NULL;
-      rc_sexp = sexp_build (&result, NULL, "(value %b)", (int)unpadlen, unpad);
-      *r_plain = sexp_null_cond (result, ct_is_not_zero (rc));
-      dummy = sexp_null_cond (result, ct_is_zero (rc));
-      sexp_release (dummy);
-      rc = ct_ulong_select (rc_sexp, rc,
-			    ct_is_zero (rc) & ct_is_not_zero (rc_sexp));
+      if (!rc)
+        rc = sexp_build (r_plain, NULL, "(value %b)", (int)unpadlen, unpad);
       break;
 
     default:
@@ -1574,13 +1470,9 @@ rsa_sign (gcry_sexp_t *r_sig, gcry_sexp_t s_data, gcry_sexp_t keyparms)
   RSA_public_key pk;
   gcry_mpi_t sig = NULL;
   gcry_mpi_t result = NULL;
-  unsigned int nbits = rsa_get_nbits (keyparms);
 
-  rc = rsa_check_keysize (nbits);
-  if (rc)
-    return rc;
-
-  _gcry_pk_util_init_encoding_ctx (&ctx, PUBKEY_OP_SIGN, nbits);
+  _gcry_pk_util_init_encoding_ctx (&ctx, PUBKEY_OP_SIGN,
+                                   rsa_get_nbits (keyparms));
 
   /* Extract the data.  */
   rc = _gcry_pk_util_data_to_mpi (s_data, &data, &ctx);
@@ -1618,7 +1510,7 @@ rsa_sign (gcry_sexp_t *r_sig, gcry_sexp_t s_data, gcry_sexp_t keyparms)
   if ((ctx.flags & PUBKEY_FLAG_NO_BLINDING))
     secret (sig, data, &sk);
   else
-    secret_blinded (sig, data, &sk, nbits);
+    secret_blinded (sig, data, &sk, ctx.nbits);
   if (DBG_CIPHER)
     log_printmpi ("rsa_sign    res", sig);
 
@@ -1680,13 +1572,9 @@ rsa_verify (gcry_sexp_t s_sig, gcry_sexp_t s_data, gcry_sexp_t keyparms)
   gcry_mpi_t data = NULL;
   RSA_public_key pk = { NULL, NULL };
   gcry_mpi_t result = NULL;
-  unsigned int nbits = rsa_get_nbits (keyparms);
 
-  rc = rsa_check_verify_keysize (nbits);
-  if (rc)
-    return rc;
-
-  _gcry_pk_util_init_encoding_ctx (&ctx, PUBKEY_OP_VERIFY, nbits);
+  _gcry_pk_util_init_encoding_ctx (&ctx, PUBKEY_OP_VERIFY,
+                                   rsa_get_nbits (keyparms));
 
   /* Extract the data.  */
   rc = _gcry_pk_util_data_to_mpi (s_data, &data, &ctx);
@@ -1694,7 +1582,7 @@ rsa_verify (gcry_sexp_t s_sig, gcry_sexp_t s_data, gcry_sexp_t keyparms)
     goto leave;
   if (DBG_CIPHER)
     log_printmpi ("rsa_verify data", data);
-  if (ctx.encoding != PUBKEY_ENC_PSS && mpi_is_opaque (data))
+  if (mpi_is_opaque (data))
     {
       rc = GPG_ERR_INV_DATA;
       goto leave;
@@ -1818,96 +1706,6 @@ compute_keygrip (gcry_md_hd_t md, gcry_sexp_t keyparam)
      Self-test section.
  */
 
-static const char *
-selftest_hash_sign_2048 (gcry_sexp_t pkey, gcry_sexp_t skey)
-{
-  int md_algo = GCRY_MD_SHA256;
-  gcry_md_hd_t hd = NULL;
-  const char *data_tmpl = "(data (flags pkcs1) (hash %s %b))";
-  static const char sample_data[] =
-    "11223344556677889900aabbccddeeff"
-    "102030405060708090a0b0c0d0f01121";
-  static const char sample_data_bad[] =
-    "11223344556677889900aabbccddeeff"
-    "802030405060708090a0b0c0d0f01121";
-
-  const char *errtxt = NULL;
-  gcry_error_t err;
-  gcry_sexp_t sig = NULL;
-  /* raw signature data reference */
-  const char ref_data[] =
-    "518f41dea3ad884e93eefff8d7ca68a6f4c30d923632e35673651d675cebd652"
-    "a44ed66f6879b18f3d48b2d235b1dd78f6189be1440352cc94231a55c1f93109"
-    "84616b2841c42fe9a6e37be34cd188207209bd028e2fa93e721fbac40c31a068"
-    "1253b312d4e07addb9c7f3d508fa89f218ea7c7f7b9f6a9b1e522c19fa1cd839"
-    "93f9d4ca2f16c3d0b9abafe5e63e848152afc72ce7ee19ea45353116f85209ea"
-    "b9de42129dbccdac8faa461e8e8cc2ae801101cc6add4ba76ccb752030b0e827"
-    "7352b11cdecebae9cdc9a626c4701cd9c85cd287618888c5fae8b4d0ba48915d"
-    "e5cc64e3aee2ba2862d04348ea71f65454f74f9fd1e3108005cc367ca41585a4";
-  gcry_mpi_t ref_mpi = NULL;
-  gcry_mpi_t sig_mpi = NULL;
-
-  err = _gcry_md_open (&hd, md_algo, 0);
-  if (err)
-    {
-      errtxt = "gcry_md_open failed";
-      goto leave;
-    }
-
-  _gcry_md_write (hd, sample_data, sizeof(sample_data));
-
-  err = _gcry_pk_sign_md (&sig, data_tmpl, hd, skey, NULL);
-  if (err)
-    {
-      errtxt = "signing failed";
-      goto leave;
-    }
-
-  err = _gcry_mpi_scan(&ref_mpi, GCRYMPI_FMT_HEX, ref_data, 0, NULL);
-  if (err)
-    {
-      errtxt = "converting ref_data to mpi failed";
-      goto leave;
-    }
-
-  err = _gcry_sexp_extract_param(sig, "sig-val!rsa", "s", &sig_mpi, NULL);
-  if (err)
-    {
-      errtxt = "extracting signature data failed";
-      goto leave;
-    }
-
-  if (mpi_cmp (sig_mpi, ref_mpi))
-    {
-      errtxt = "signature does not match reference data";
-      goto leave;
-    }
-
-  err = _gcry_pk_verify_md (sig, data_tmpl, hd, pkey, NULL);
-  if (err)
-    {
-      errtxt = "verify failed";
-      goto leave;
-    }
-
-  _gcry_md_reset(hd);
-  _gcry_md_write (hd, sample_data_bad, sizeof(sample_data_bad));
-  err = _gcry_pk_verify_md (sig, data_tmpl, hd, pkey, NULL);
-  if (gcry_err_code (err) != GPG_ERR_BAD_SIGNATURE)
-    {
-      errtxt = "bad signature not detected";
-      goto leave;
-    }
-
-
- leave:
-  sexp_release (sig);
-  _gcry_md_close (hd);
-  _gcry_mpi_release (ref_mpi);
-  _gcry_mpi_release (sig_mpi);
-  return errtxt;
-}
-
 static const char *
 selftest_sign_2048 (gcry_sexp_t pkey, gcry_sexp_t skey)
 {
@@ -2086,7 +1884,7 @@ selftest_encr_2048 (gcry_sexp_t pkey, gcry_sexp_t skey)
   ciphertext = extract_a_from_sexp (encr);
   if (!ciphertext)
     {
-      errtxt = "gcry_pk_encrypt returned garbage";
+      errtxt = "gcry_pk_decrypt returned garbage";
       goto leave;
     }
 
@@ -2144,7 +1942,7 @@ selftest_encr_2048 (gcry_sexp_t pkey, gcry_sexp_t skey)
 
 
 static gpg_err_code_t
-selftests_rsa (selftest_report_func_t report, int extended)
+selftests_rsa (selftest_report_func_t report)
 {
   const char *what;
   const char *errtxt;
@@ -2172,26 +1970,15 @@ selftests_rsa (selftest_report_func_t report, int extended)
       goto failed;
     }
 
-  if (extended)
-    {
-      what = "sign";
-      errtxt = selftest_sign_2048 (pkey, skey);
-      if (errtxt)
-        goto failed;
-    }
-
-  what = "digest sign";
-  errtxt = selftest_hash_sign_2048 (pkey, skey);
+  what = "sign";
+  errtxt = selftest_sign_2048 (pkey, skey);
   if (errtxt)
     goto failed;
 
-  if (extended)
-    {
-      what = "encrypt";
-      errtxt = selftest_encr_2048 (pkey, skey);
-      if (errtxt)
-        goto failed;
-    }
+  what = "encrypt";
+  errtxt = selftest_encr_2048 (pkey, skey);
+  if (errtxt)
+    goto failed;
 
   sexp_release (pkey);
   sexp_release (skey);
@@ -2212,10 +1999,12 @@ run_selftests (int algo, int extended, selftest_report_func_t report)
 {
   gpg_err_code_t ec;
 
+  (void)extended;
+
   switch (algo)
     {
     case GCRY_PK_RSA:
-      ec = selftests_rsa (report, extended);
+      ec = selftests_rsa (report);
       break;
     default:
       ec = GPG_ERR_PUBKEY_ALGO;
