@@ -188,9 +188,11 @@
 #define DRBG_HASHSHA1		((u32)1<<4)
 #define DRBG_HASHSHA224		((u32)1<<5)
 #define DRBG_HASHSHA256		((u32)1<<6)
+#define DRBG_HASHSHA384		((u32)1<<7)
 #define DRBG_HASHSHA512		((u32)1<<8)
 #define DRBG_HASH_MASK		(DRBG_HASHSHA1 | DRBG_HASHSHA224 \
-				 | DRBG_HASHSHA256 | DRBG_HASHSHA512)
+				 | DRBG_HASHSHA256 | DRBG_HASHSHA384 \
+				 | DRBG_HASHSHA512)
 /* type modifiers (A.3)*/
 #define DRBG_HMAC		((u32)1<<12)
 #define DRBG_SYM128		((u32)1<<13)
@@ -209,18 +211,23 @@
 #define DRBG_NOPR_CTRAES256 (DRBG_CTRAES | DRBG_SYM256)
 #define DRBG_PR_HASHSHA1     (DRBG_PREDICTION_RESIST | DRBG_HASHSHA1)
 #define DRBG_PR_HASHSHA256   (DRBG_PREDICTION_RESIST | DRBG_HASHSHA256)
+#define DRBG_PR_HASHSHA384   (DRBG_PREDICTION_RESIST | DRBG_HASHSHA384)
 #define DRBG_PR_HASHSHA512   (DRBG_PREDICTION_RESIST | DRBG_HASHSHA512)
 #define DRBG_NOPR_HASHSHA1   (DRBG_HASHSHA1)
 #define DRBG_NOPR_HASHSHA256 (DRBG_HASHSHA256)
+#define DRBG_NOPR_HASHSHA384 (DRBG_HASHSHA384)
 #define DRBG_NOPR_HASHSHA512 (DRBG_HASHSHA512)
 #define DRBG_PR_HMACSHA1     (DRBG_PREDICTION_RESIST | DRBG_HASHSHA1 \
                               | DRBG_HMAC)
 #define DRBG_PR_HMACSHA256   (DRBG_PREDICTION_RESIST | DRBG_HASHSHA256 \
                               | DRBG_HMAC)
+#define DRBG_PR_HMACSHA384   (DRBG_PREDICTION_RESIST | DRBG_HASHSHA384 \
+                              | DRBG_HMAC)
 #define DRBG_PR_HMACSHA512   (DRBG_PREDICTION_RESIST | DRBG_HASHSHA512 \
                               | DRBG_HMAC)
 #define DRBG_NOPR_HMACSHA1   (DRBG_HASHSHA1 | DRBG_HMAC)
 #define DRBG_NOPR_HMACSHA256 (DRBG_HASHSHA256 | DRBG_HMAC)
+#define DRBG_NOPR_HMACSHA384 (DRBG_HASHSHA384 | DRBG_HMAC)
 #define DRBG_NOPR_HMACSHA512 (DRBG_HASHSHA512 | DRBG_HMAC)
 
 
@@ -291,7 +298,7 @@ struct drbg_state_ops_s
 struct drbg_test_data_s
 {
   drbg_string_t *testentropy;	/* TEST PARAMETER: test entropy */
-  unsigned int fail_seed_source:1; /* If set, the seed function will
+  int fail_seed_source:1;	/* If set, the seed function will
                                  * return an error. */
 };
 
@@ -308,8 +315,8 @@ struct drbg_state_s
 				 * operation -- allocated during init */
   void *priv_data;		/* Cipher handle */
   gcry_cipher_hd_t ctr_handle;	/* CTR mode cipher handle */
-  unsigned int seeded:1;	/* DRBG fully seeded? */
-  unsigned int pr:1;		/* Prediction resistance enabled? */
+  int seeded:1;			/* DRBG fully seeded? */
+  int pr:1;			/* Prediction resistance enabled? */
   /* Taken from libgcrypt ANSI X9.31 DRNG: We need to keep track of the
    * process which did the initialization so that we can detect a fork.
    * The volatile modifier is required so that the compiler does not
@@ -334,9 +341,6 @@ enum drbg_prefixes
  * Global variables
  ***************************************************************/
 
-/* The instance of the DRBG, to be refereed by drbg_state.  */
-static struct drbg_state_s drbg_instance;
-
 /* Global state variable holding the current instance of the DRBG.  */
 static drbg_state_t drbg_state;
 
@@ -352,10 +356,12 @@ static const struct drbg_core_s drbg_cores[] = {
   /* Hash DRBGs */
   {DRBG_HASHSHA1, 55, 20, GCRY_MD_SHA1},
   {DRBG_HASHSHA256, 55, 32, GCRY_MD_SHA256},
+  {DRBG_HASHSHA384, 111, 48, GCRY_MD_SHA384},
   {DRBG_HASHSHA512, 111, 64, GCRY_MD_SHA512},
   /* HMAC DRBGs */
   {DRBG_HASHSHA1   | DRBG_HMAC, 20, 20, GCRY_MD_SHA1},
   {DRBG_HASHSHA256 | DRBG_HMAC, 32, 32, GCRY_MD_SHA256},
+  {DRBG_HASHSHA384 | DRBG_HMAC, 48, 48, GCRY_MD_SHA384},
   {DRBG_HASHSHA512 | DRBG_HMAC, 64, 64, GCRY_MD_SHA512},
   /* block ciphers */
   {DRBG_CTRAES | DRBG_SYM128, 32, 16, GCRY_CIPHER_AES128},
@@ -534,7 +540,7 @@ drbg_sec_strength (u32 flags)
   else if (flags & DRBG_SYM192)
     return 24;
   else if ((flags & DRBG_SYM256) || (flags & DRBG_HASHSHA256) ||
-	   (flags & DRBG_HASHSHA512))
+	   (flags & DRBG_HASHSHA384) || (flags & DRBG_HASHSHA512))
     return 32;
   else
     return 32;
@@ -612,12 +618,9 @@ drbg_get_entropy (drbg_state_t drbg, unsigned char *buffer,
   read_cb_buffer = buffer;
   read_cb_size = len;
   read_cb_len = 0;
-#if USE_RNDGETENTROPY
-  rc = _gcry_rndgetentropy_gather_random (drbg_read_cb, 0, len,
-                                          GCRY_VERY_STRONG_RANDOM);
-#elif USE_RNDOLDLINUX
-  rc = _gcry_rndoldlinux_gather_random (drbg_read_cb, 0, len,
-                                        GCRY_VERY_STRONG_RANDOM);
+#if USE_RNDLINUX
+  rc = _gcry_rndlinux_gather_random (drbg_read_cb, 0, len,
+				     GCRY_VERY_STRONG_RANDOM);
 #elif USE_RNDUNIX
   rc = _gcry_rndunix_gather_random (drbg_read_cb, 0, len,
 				    GCRY_VERY_STRONG_RANDOM);
@@ -1777,7 +1780,9 @@ _drbg_init_internal (u32 flags, drbg_string_t *pers)
     }
   else
     {
-      drbg_state = &drbg_instance;
+      drbg_state = xtrycalloc_secure (1, sizeof *drbg_state);
+      if (!drbg_state)
+	return gpg_err_code_from_syserror ();
     }
   if (flags & DRBG_PREDICTION_RESIST)
     pr = 1;
@@ -1855,25 +1860,16 @@ _gcry_rngdrbg_reinit (const char *flagstr, gcry_buffer_t *pers, int npers)
   return ret;
 }
 
-/* Release resources used by this DRBG module.  That is, close the FDs
- * of the random gather module (if any), and release memory used.
- */
+/* Try to close the FDs of the random gather module.  This is
+ * currently only implemented for rndlinux. */
 void
 _gcry_rngdrbg_close_fds (void)
 {
+#if USE_RNDLINUX
   drbg_lock ();
-#if USE_RNDGETENTROPY
-  _gcry_rndgetentropy_gather_random (NULL, 0, 0, 0);
-#endif
-#if USE_RNDOLDLINUX
-  _gcry_rndoldlinux_gather_random (NULL, 0, 0, 0);
-#endif
-  if (drbg_state)
-    {
-      drbg_uninstantiate (drbg_state);
-      drbg_state = NULL;
-    }
+  _gcry_rndlinux_gather_random (NULL, 0, 0, 0);
   drbg_unlock ();
+#endif
 }
 
 /* Print some statistics about the RNG.  */
@@ -1931,9 +1927,6 @@ _gcry_rngdrbg_randomize (void *buffer, size_t length,
    * key, either a re-init or a reseed is sufficient for a fork */
   if (drbg_state->seed_init_pid != getpid ())
     {
-      /* Update the PID recorded.  */
-      drbg_state->seed_init_pid = getpid ();
-
       /* We are in a child of us. Perform a reseeding. */
       if (drbg_reseed (drbg_state, NULL))
 	{
