@@ -34,7 +34,6 @@
 #include "t-common.h"
 #define N_TESTS 9
 
-static int in_fips_mode;
 
 static void
 print_mpi (const char *text, gcry_mpi_t a)
@@ -180,18 +179,8 @@ test_cv_hl (int testno, const char *k_str, const char *u_str,
   xfree (buffer);
   buffer = NULL;
 
-  err = gcry_pk_encrypt (&s_result, s_data, s_pk);
-  if (in_fips_mode)
-    {
-      if (!err)
-        fail ("gcry_pk_encrypt is not expected to work in FIPS mode for test %d",
-              testno);
-      if (verbose > 1)
-        info ("not executed in FIPS mode\n");
-      goto leave;
-    }
-  if (err)
-    fail ("gcry_pk_encrypt goto leavefailed for test %d: %s", testno,
+  if ((err = gcry_pk_encrypt (&s_result, s_data, s_pk)))
+    fail ("gcry_pk_encrypt failed for test %d: %s", testno,
           gpg_strerror (err));
 
   s_tmp = gcry_sexp_find_token (s_result, "s", 0);
@@ -268,17 +257,7 @@ test_cv_x448 (int testno, const char *k_str, const char *u_str,
       goto leave;
     }
 
-  err = gcry_ecc_mul_point (GCRY_ECC_CURVE448, result, scalar, point);
-  if (in_fips_mode)
-    {
-      if (err != GPG_ERR_NOT_SUPPORTED)
-        fail ("gcry_ecc_mul_point is not expected to work in FIPS mode for test %d: %s",
-              testno, gpg_strerror (err));
-      if (verbose > 1)
-        info ("not executed in FIPS mode\n");
-      goto leave;
-    }
-  if (err)
+  if ((err = gcry_ecc_mul_point (GCRY_ECC_CURVE448, result, scalar, point)))
     fail ("gcry_ecc_mul_point failed for test %d: %s", testno,
           gpg_strerror (err));
 
@@ -317,14 +296,14 @@ test_cv (int testno, const char *k_str, const char *u_str,
 static void
 test_it (int testno, const char *k_str, int iter, const char *result_str)
 {
-  gcry_ctx_t ctx = NULL;
+  gcry_ctx_t ctx;
   gpg_error_t err;
   void *buffer = NULL;
   size_t buflen;
   gcry_mpi_t mpi_k = NULL;
   gcry_mpi_t mpi_x = NULL;
   gcry_mpi_point_t P = NULL;
-  gcry_mpi_point_t Q = NULL;
+  gcry_mpi_point_t Q;
   int i;
   gcry_mpi_t mpi_kk = NULL;
 
@@ -332,15 +311,6 @@ test_it (int testno, const char *k_str, int iter, const char *result_str)
     info ("Running test %d: iteration=%d\n", testno, iter);
 
   gcry_mpi_ec_new (&ctx, NULL, "X448");
-  if (in_fips_mode)
-    {
-      if (ctx)
-        fail ("gcry_mpi_ec_new should fail in FIPS mode for test %d",
-              testno);
-      if (verbose > 1)
-        info ("not executed in FIPS mode\n");
-      goto leave;
-    }
   Q = gcry_mpi_point_new (0);
 
   if (!(buffer = hex2buffer (k_str, &buflen)) || buflen != 56)
@@ -613,9 +583,6 @@ main (int argc, char **argv)
   xgcry_control ((GCRYCTL_ENABLE_QUICK_RANDOM, 0));
   xgcry_control ((GCRYCTL_INITIALIZATION_FINISHED, 0));
 
-  if (gcry_fips_mode_active ())
-    in_fips_mode = 1;
-
   start_timer ();
   check_x448 ();
   stop_timer ();
